/* PlayBack.c
 *
 * Kevin O'Connor 9/22/97
 *
 * Routines neccessary to playback a game recording.
 *
 * $Log: playback.c,v $
 * Revision 1.1.1.1  1998/11/01 17:24:11  siegl
 * COW 3.0 initial revision
 * */
#include <setjmp.h>
#include "config.h"

#include INC_MACHINE_ENDIAN

#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include INC_SYS_TIME
#include INC_SYS_WAIT
#include INC_SYS_RESOURCE
#include INC_SYS_SELECT
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <pwd.h>
#include <string.h>
#include <ctype.h>
#include "Wlib.h"
#include "defs.h"
#include "struct.h"
#include "data.h"
#include "packets.h"
#include "version.h"
#include "patchlevel.h"
#include "cowapi.h"
#include "map.h"
#include "spopt.h"
#include "defs.h"
#include "playerlist.h"

#ifdef RECORDGAME
extern int opened_info;				 /* counter for infowin *

						  * 
						  * * popup, 6/1/93 LAB */
#define RETURNBASE 10

extern jmp_buf env;

int     pbdelay = 200000;

struct player dummyme;

int
        pbmain(char *name)
{
  int     i;
  int     s_type;

  playback = PL_FORWARD;
  pseudo[0] = defpasswd[0] = '\0';

  i = setjmp(env);				 /* Error while initializing */
  if (i >= RETURNBASE)
    return (i - RETURNBASE);			 /* Terminate with retcode */

  for (i = 0; i < 80; i++)
    {
      outmessage[i] = '\0';
    }

  SRANDOM(time(0));

  initDefaults(deffile);

  SRANDOM(getpid() * time((LONG *) 0));

  newwin(display_host, name);

  resetdefaults();
  if (censorMessages)
    initCensoring();

  /* open memory...? */
  openmem();

  recordFile = fopen(recordFileName, "rb");
  if (recordFile == NULL)
    {
      perror(recordFileName);
      return (1);

    }

  me = &dummyme;
  myship = &(me->p_ship);
  mystats = &(me->p_stats);

  me->p_x = me->p_y = 50000;
  getship(myship, CRUISER);
  shipchange(CRUISER);

  /* Get first packet from file */
  readFromFile();

  lastm = mctl->mc_current;

  mapAll();

  (void) SIGNAL(SIGINT, SIG_IGN);

  /* Instructions from getname() */
  MZERO(mystats, sizeof(struct stats));

  mystats->st_tticks = 1;
  for (i = 0; i < 95; i++)
    {
      mystats->st_keymap[i] = i + 32;
      mystats->st_keymap[i + 96] = i + 32 + 96;

#ifdef MOUSE_AS_SHIFT
      mystats->st_keymap[i + 192] = i + 32;
      mystats->st_keymap[i + 288] = i + 32;
      mystats->st_keymap[i + 384] = i + 32;
#endif
    }
  mystats->st_keymap[95] = 0;
  mystats->st_flags = ST_MAPMODE + ST_NAMEMODE + ST_SHOWSHIELDS +
      ST_KEEPPEACE + ST_SHOWLOCAL * 2 + ST_SHOWGLOBAL * 2;

  /* End getname() */

  phaserWindow = booleanDefault("phaserWindow", phaserWindow);


#ifdef AUTOKEY
  /* autokey.c */
  autoKeyDefaults();
#endif /* AUTOKEY */

  initkeymap();

  /* Set p_hostile to hostile, so if keeppeace is on, the guy starts off * *
   * hating everyone (like a good fighter should) */
  me->p_hostile = (FED | ROM | KLI | ORI);

  if (!newDashboard)
    {
      char    buf[128];

      sprintf(buf,
	   "Maximum:      %2d  %3d %3d               %3d   %6d   %3d   %3d",
	      0, 0, 0, 0, 0, 0, 0);
      W_WriteText(tstatw, 50, 27, textColor, buf, strlen(buf), W_RegularFont);
    }

#ifdef AUTOKEY
  if (autoKey)
    {
      /* XX: changes entire state of display */
      W_AutoRepeatOff();
    }
#endif

#ifdef SOUND
  Init_Sound();
#endif

  i = setjmp(env);				 /* Reentry point of game */
  if (i >= RETURNBASE)
    return (i - RETURNBASE);			 /* Terminate with retcode */

#ifdef SOUND
  Abort_Sound(ENGINE_SOUND);
#endif

#ifdef nodef
  /* Code from entrywindow() */
  {
    run_clock(time(0));

    if (remap[me->p_team] == NOBODY)
      RedrawPlayerList();			 /* When you first login */
    else
      UpdatePlayerList();			 /* Otherwise */

  }
  /* End entrywindow() */
#endif

#ifdef BRMH
  redrawall = 2;
#else
  redrawall = 1;
#endif

  getship(myship, myship->s_type);
  shipchange(myship->s_type);
  enter();
  calibrate_stats();
  W_ClearWindow(w);

  me->p_status = PALIVE;			 /* Put player in game */
  me->p_ghostbuster = 0;
  PlistNoteUpdate(me->p_no);

  if (showStats)				 /* Default showstats are on. 
						  * 
						  */
    W_MapWindow(statwin);

  if (W_IsMapped(lMeter))
    redrawLMeter();

  if (W_IsMapped(pStats))			 /* support ping stuff */
    redrawPStats();

#ifdef SOUND
  Play_Sound(ENTER_SHIP_SOUND);
  Play_Sound(ENGINE_SOUND);
#endif

#ifdef HOCKEY_LINES
  init_hockey_lines();
#endif

  while (1)
    {

#ifdef nodef
      fd_set  readfds;
      struct timeval timeout;
      int     xsock = W_Socket();

      timeout.tv_sec = 0;
      timeout.tv_usec = 0;
#endif

      if (keepInfo > 0 && opened_info != -2 &&	 /* 6/1/93 LAB */
	  opened_info < 0 && infomapped)
	destroyInfo();

      if (keepInfo > 0 && opened_info != -2)
	opened_info--;

#ifdef nodef
      FD_ZERO(&readfds);
      FD_SET(xsock, &readfds);

      if (select(max_fd, &readfds, 0, 0, &timeout))
	{
	  if (FD_ISSET(xsock, &readfds))
	    {
	      process_event();
	      /* NOTE: we're no longer calling XPending(), need this */
	      W_Flush();
	    }
	}
#else
      if (W_EventsQueuedCk())
	{
	  process_event();
	  /* W_Flush(); */
	}
#endif

      {
	intrupt();
	W_Flush();
	usleep(pbdelay);
      }
    }
}

void
        pbsetspeed(char key)
{
  if (playback == PL_PAUSE)
    playback = PL_FORWARD;
  switch (key)
    {
    case '0':
      playback = PL_PAUSE;
      break;
    case '1':
      pbdelay = 800000;
      break;
    case '2':
      pbdelay = 400000;
      break;
    case '3':
      pbdelay = 200000;
      break;
    case '4':
      pbdelay = 100000;
      break;
    case '5':
      pbdelay = 50000;
      break;
    case '6':
      pbdelay = 25000;
      break;
    case '7':
      pbdelay = 12500;
      break;
    case '8':
      pbdelay = 6250;
      break;
    case '9':
    case '#':
    case '!':
    case '@':
    case '%':
      pbdelay = 0;
      break;
    case '<':
      pbdelay /= 2;
      break;
    case '>':
      pbdelay *= 2;
      break;
    case 'R':
      {
	fseek(recordFile, 0, SEEK_SET);
      }
    case '(':
      /* playback = PL_REVERSE; */
      break;
    case ')':
      playback = PL_FORWARD;
      break;
    }
}

void
        pblockplayer(int who)
{
  /* F_many_self = 1; */
  me = &players[who];
}

void
        pblockplanet(int pl)
{
  /* F_many_self = 1; */
  me = &dummyme;
  me->p_x = planets[pl].pl_x;
  me->p_y = planets[pl].pl_y;
}

inline int
        ckRecordPacket(char packet)
{
  switch (packet)
    {
    case SP_MESSAGE:
    case SP_PLAYER_INFO:
    case SP_KILLS:
    case SP_PLAYER:
    case SP_TORP_INFO:
    case SP_TORP:
    case SP_PHASER:
    case SP_PLASMA_INFO:
    case SP_PLASMA:
    case SP_WARNING:
      /* SP_MOTD */
    case SP_YOU:
    case SP_STATUS:
    case SP_PLANET:
    case SP_FLAGS:
    case SP_MASK:
    case SP_PSTATUS:
    case SP_HOSTILE:
    case SP_STATS:
    case SP_PL_LOGIN:
    case SP_PLANET_LOC:
    case SP_SEQUENCE:
    case SP_SC_SEQUENCE:
    case SP_S_MESSAGE:
    case SP_S_WARNING:
    case SP_S_YOU:
    case SP_S_YOU_SS:
    case SP_S_PLAYER:
    case SP_SHIP_CAP:
    case SP_S_TORP:
    case SP_S_TORP_INFO:
    case SP_S_8_TORP:
    case SP_S_PLANET:
    case SP_S_SEQUENCE:
    case SP_S_PHASER:
    case SP_S_KILLS:
    case SP_S_STATS:

      return 1;
    }
  return 0;
}

extern struct packet_handler handlers[];

int
        readFromFile()
{
#define MAXPACKETSIZE 128
  static uint aligned_buf[MAXPACKETSIZE / sizeof(uint)];
  char   *buf = (char *) &aligned_buf;

  if (playback == PL_PAUSE)
    return 0;

  if (playback == PL_FORWARD)
    {
      while (1)
	{
	  int     size, count, loc;

	  count = fread(buf, 1, 4, recordFile);
	  if (count < 4)
	    {
	      playback == PL_PAUSE;
	      return 0;
	    }

	  loc = 4;
	  size = handlers[buf[0]].size;
	  if (size == -1)
	    {
	      if (buf[0] == SP_S_MESSAGE)
		{
		  /* UGH.  SP_S_MESSAGE needs next word to calculate size */
		  count = fread(buf + loc, 1, 4, recordFile);
		  if (count < 4)
		    {
		      playback == PL_PAUSE;
		      return 0;
		    }
		  loc += 4;
		}
	      size = getvpsize(buf);
	    }
	  if (size > loc)
	    count = fread(buf + loc, 1, size - loc, recordFile);
	  if (debug)
	    printf("Reading packet %d\n", buf[0]);
	  if (count < size - loc)
	    {
	      playback == PL_PAUSE;
	      return 0;
	    }

	  (*(handlers[buf[0]].handler)) (buf

#ifdef CORRUPTED_PACKETS
					 ,recordFile
#endif

	      );

	  if (buf[0] == SP_SEQUENCE || buf[0] == SP_SC_SEQUENCE
	      || buf[0] == SP_S_SEQUENCE)
	    return 1;
	}
    }
}
#endif /* RECORDGAME */
